# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
import json
import collections
from .exceptions import ValidationError, ModelValidationError
from .model import get_type_specifier_name, get_model_class_for_type, get_type_name_for_model, Model
from .types import ModelType, StringType, BooleanType, IntType


class FieldSerializer():
    @abstractmethod
    def serialize_field(self, value, model_type, field_name, field, **kwargs):
        raise TypeError('Don\'t know how to serialize {}'.format(field.type))  # pragma: no cover

    @abstractmethod
    def deserialize_field(self, value, model_type, field_name, field, **kwargs):
        raise TypeError('Don\'t know how to deserialize {}'.format(field.type))  # pragma: no cover


class VerbatimFieldSerializer(FieldSerializer):
    def serialize_field(self, value, model_type, field_name, field, **kwargs):
        return value

    def deserialize_field(self, value, model_type, field_name, field, **kwargs):
        return value


def make_field_filter(fields):
    if fields is None:
        return lambda x: True
    if isinstance(fields, collections.Container):
        return lambda x: x[0] in fields
    else:
        return fields


def iter_model_fields(model_class, fields=None, **kwargs):
    items = model_class.fields.items()
    if fields is None:
        return items

    return filter(make_field_filter(fields), items)


class ModelSerializer(metaclass=ABCMeta):
    def __init__(self, field_serializer):
        self.field_serializer = field_serializer

    def _iter_model_fields(self, model_class, **kwargs):
        return iter_model_fields(model_class, **kwargs)

    @abstractmethod
    def _serialize_model(self, value, model_type, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    def serialize_model(self, value, model_type=None, **kwargs):
        if not isinstance(value, Model):
            raise TypeError('Value is not an instance of Model')
        if model_type is None:
            model_type = type(value)
        return self._serialize_model(value, model_type)

    @abstractmethod
    def _deserialize_model(self, value, model_type, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    def deserialize_model(self, value, model_or_model_type, **kwargs):
        try:
            return self._deserialize_model(value, model_or_model_type, **kwargs)
        except ModelValidationError:
            raise  # pragma: no cover
        except ValidationError as e:
            mve = ModelValidationError()
            mve.__cause__ = e
            mve.add_error(e)
            raise mve


class DictModelSerializer(ModelSerializer):
    def _serialize_model(self, value, model_type, **kwargs):
        if value is None:
            return None

        model_class = type(value)

        result = {}
        for name, field in self._iter_model_fields(model_class, **kwargs):
            if isinstance(field.type, ModelType):
                result[name] = self._serialize_model(getattr(value, name), model_class, **kwargs)
            else:
                result[name] = self.field_serializer.serialize_field(getattr(value, name), model_class, name, field,
                                                                     **kwargs)

        type_specifier_name = get_type_specifier_name(model_type)
        if type_specifier_name:
            model_type_name = get_type_name_for_model(model_type, model_class)
            result[type_specifier_name] = model_type_name

        return result

    def _deserialize_model(self, value, model_or_model_type, **kwargs):
        if value is None:
            return None

        if not isinstance(value, dict):
            raise ValidationError('Model deserialization requires dictionary type')

        if isinstance(model_or_model_type, Model):
            model = model_or_model_type
            model_class = type(model)
        else:
            type_name = None
            type_specifier_name = get_type_specifier_name(model_or_model_type)
            if type_specifier_name is not None:
                if type_specifier_name not in value:
                    raise ValidationError('Polymorphic model requires type specifier')
                type_name = value[type_specifier_name]

            model_class = get_model_class_for_type(model_or_model_type, type_name)
            model = model_class()

        for name, field in self._iter_model_fields(model_class, **kwargs):
            if isinstance(field.type, ModelType):
                deserialized_value = self._deserialize_model(value.get(name), field.type.native_type, **kwargs)
            else:
                deserialized_value = self.field_serializer.deserialize_field(value.get(name), model_class, name, field,
                                                                             **kwargs)
            setattr(model, name, deserialized_value)

        return model


class JsonModelSerializer(DictModelSerializer):
    def __init__(self, field_serializer, sort_keys=False):
        super().__init__(field_serializer)
        self.sort_keys = sort_keys

    def serialize_model(self, value, model_type=None, **kwargs):
        return json.dumps(super().serialize_model(value, model_type, **kwargs), sort_keys=self.sort_keys)

    def _deserialize_model(self, value, model_or_model_type, **kwargs):
        try:
            value = json.loads(value)
        except ValueError as e:
            raise ValidationError('Value is not a valid JSON: ' + str(e))
        return super()._deserialize_model(value, model_or_model_type, **kwargs)


# -*- coding: utf-8 -*-
from collections.abc import Iterable
from datetime import datetime, date, time
from importlib import import_module
from urllib.parse import urlparse
import re
import builtins
from .exceptions import ValidationError


class ValueType:
    def __init__(self, validators=None):
        if validators is None:
            self.validators = []
        else:
            self.validators = validators

    def validate(self, value):
        if not isinstance(value, self.native_type):
            raise ValidationError('{!r} is not an instance of {}'.format(value, self.native_type))
        for validator in self.validators:
            validator(value)

    @property
    def native_type(self):
        return object

    @property
    def default_value(self):
        return self.native_type()


class BooleanType(ValueType):
    @property
    def native_type(self):
        return bool


class SizedType(ValueType):
    def __init__(self, min_length=None, max_length=None, **kwargs):
        super().__init__(**kwargs)
        self.min_length = min_length
        self.max_length = max_length

    def validate(self, value):
        super().validate(value)
        if self.min_length is not None and len(value) < self.min_length:
            raise ValidationError('{!r} is too short, minimal allowed length is {} characters'
                                  .format(value, self.min_length))
        if self.max_length is not None and len(value) > self.max_length:
            raise ValidationError('{!r} is too long, maximal allowed length is {} characters'
                                  .format(value, self.max_length))


class ComparableType(ValueType):
    def __init__(self, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def validate(self, value):
        super().validate(value)
        if self.min_value is not None and value < self.min_value:
            raise ValidationError('{!r} is too small, minimal allowed value is {!r}'.format(value, self.min_value))
        if self.max_value is not None and value > self.max_value:
            raise ValidationError('{!r} is too small, minimal allowed value is {!r}'.format(value, self.max_value))


class StringType(SizedType):
    def __init__(self, regex=None, **kwargs):
        super().__init__(**kwargs)
        if regex is None:
            self.regex = None
        elif isinstance(regex, str):
            self.regex = re.compile(regex)
        else:
            self.regex = regex

    @property
    def native_type(self):
        return str

    def validate(self, value):
        super().validate(value)
        if self.regex is not None and not self.regex.search(value):
            raise ValidationError('{!r} does not match validation pattern')


class UrlType(StringType):
    def __init__(self, scheme=None, **kwargs):
        super().__init__(**kwargs)
        if scheme is None:
            self.scheme = None
        elif isinstance(scheme, str):
            self.scheme = (scheme,)
        elif isinstance(scheme, Iterable):
            self.scheme = tuple(scheme)
        else:
            raise TypeError('Invalid value type {} for scheme constraint'.format(type(scheme)))

    def validate(self, value):
        super().validate(value)
        parsed = urlparse(value)
        if self.scheme is not None and parsed.scheme not in self.scheme:
            raise ValidationError('{!r} scheme is not {!r}'.format(value, self.scheme))


class IntType(ComparableType):
    @property
    def native_type(self):
        return int


class IterableType(SizedType):
    def __init__(self, item_type=None, **kwargs):
        super().__init__(**kwargs)
        self.item_type = item_type

    def validate(self, value):
        super().validate(value)
        if self.item_type is not None:
            for item in value:
                self.item_type.validate(item)


class ListType(IterableType):
    @property
    def native_type(self):
        return list


class SetType(IterableType):
    @property
    def native_type(self):
        return set


class DictType(SizedType):
    def __init__(self, key_type=None, value_type=None, **kwargs):
        super().__init__(**kwargs)
        self.key_type = key_type
        self.value_type = value_type

    @property
    def native_type(self):
        return dict

    def validate(self, value):
        super().validate(value)
        if self.key_type is not None or self.value_type is not None:
            for item_key, item_value in value.items():
                if self.key_type is not None:
                    self.key_type.validate(item_key)
                if self.value_type is not None:
                    self.value_type.validate(item_value)


def import_object(fully_qualified_name):
    parts = [part for part in fully_qualified_name.split('.') if part]

    module = None
    module_parts = 0
    part_count = len(parts)
    while module_parts < part_count:
        try:
            part_module = import_module('.'.join(parts[:module_parts+1]))
        except ImportError:
            break
        module_parts += 1
        module = part_module

    if module is None:
        module = builtins

    value = module
    for part in parts[module_parts:]:
        value = getattr(value, part)

    return value


class ModelType(ValueType):
    def __init__(self, model_class, **kwargs):
        super().__init__(**kwargs)
        self._model_class = model_class

    @property
    def model_class(self):
        if isinstance(self._model_class, str):
            self._model_class = import_object(self._model_class)
        return self._model_class

    @property
    def native_type(self):
        return self.model_class

    def validate(self, value):
        super().validate(value)
        if value is not None:
            value.validate()


class DateTimeType(ComparableType):
    @property
    def native_type(self):
        return datetime

    @property
    def default_value(self):
        return datetime.now()


class DateType(ComparableType):
    @property
    def native_type(self):
        return date

    @property
    def default_value(self):
        return date.today()


class TimeType(ComparableType):
    @property
    def native_type(self):
        return time

    @property
    def default_value(self):
        return datetime.now().time()


# -*- coding: utf-8 -*-
from collections import OrderedDict
from .exceptions import ValidationError, ModelValidationError
from abc import ABCMeta

type_default = object()


class Field:
    def __init__(self, type, required=True, default=type_default, **kwargs):
        self.required = required
        self.default = default
        self.type = type
        for name, value in kwargs.items():
            setattr(self, name, value)

    def create_default_value(self):
        if self.default is type_default:
            if not self.required:
                return None
            return self.type.default_value
        elif callable(self.default):
            return self.default()
        else:
            return self.default

    def validate(self, value):
        if value is None:
            if self.required:
                raise ValidationError('value is required')
        else:
            self.type.validate(value)


class ModelMeta(type):
    @classmethod
    def __prepare__(mcls, name, bases):
        return OrderedDict()

    def __new__(mcls, name, bases, namespace):
        declared_fields = OrderedDict()
        new_namespace = {}
        for key, value in namespace.items():
            if isinstance(value, Field):
                declared_fields[key] = value
            else:
                new_namespace[key] = value
        new_namespace['declared_fields'] = declared_fields

        cls = super().__new__(mcls, name, bases, new_namespace)

        merged_fields = OrderedDict()
        for part_cls in (part_cls for part_cls in reversed(cls.mro()) if hasattr(part_cls, 'declared_fields')):
            for key, field in part_cls.declared_fields.items():
                merged_fields[key] = field

        cls.fields = merged_fields

        return cls


class Model(metaclass=ModelMeta):
    fields = None

    def __init__(self, **kwargs):
        for name, field in self.fields.items():
            if name in kwargs:
                setattr(self, name, kwargs[name])
            else:
                setattr(self, name, field.create_default_value())

    def validate(self):
        error = ModelValidationError()
        for name, field in self.fields.items():
            try:
                field.validate(getattr(self, name))
            except ValidationError as field_error:
                error.add_field_error(name, field_error)
        if error:
            raise error

    def __eq__(self, other):
        if type(self) != type(other):
            return NotImplemented
        for name in self.fields.keys():
            if not (getattr(self, name) == getattr(other, name)):
                return False
        return True

    def __ne__(self, other):
        if type(self) != type(other):
            return NotImplemented
        for name in self.fields.keys():
            if getattr(self, name) != getattr(other, name):
                return True
        return False

    def __repr__(self):  # pragma: no cover
        field_descr = ', '.join('{}={!r}'.format(name, getattr(self, name)) for name in type(self).fields.keys())
        return '{}({})'.format(type(self).__qualname__, field_descr)


class PolymorphicModelMeta(ABCMeta):
    def __new__(mcs, name, bases, namespace):
        for klass in bases:
            if isinstance(klass, PolymorphicModelMeta) and klass is not PolymorphicModel:
                raise TypeError('Polymorphic models cannot be subclassed')

        cls = super().__new__(mcs, name, bases, namespace)
        cls.types_to_model_classes_mro = list(cls.types_to_model_classes.items())
        cls.types_to_model_classes_mro.sort(key=lambda x: len(x[1].mro()), reverse=True)
        return cls

    def __call__(self, type_name, *args, **kwargs):
        model_class = self.get_model_class_for_type(type_name)
        if model_class is None:
            raise ValueError('Unknown model type name {!r} for class {}'.format(type_name, self))
        return model_class(*args, **kwargs)


class PolymorphicModel(metaclass=PolymorphicModelMeta):
    types_to_model_classes = {}
    types_to_model_classes_mro = []
    type_specifier_name = 'type'

    @classmethod
    def get_model_type_for_class(cls, unknown_class):
        for type_name, model_class in cls.types_to_model_classes_mro:
            if issubclass(unknown_class, model_class):
                return type_name
        return None

    @classmethod
    def get_model_class_for_type(cls, type_name):
        return cls.types_to_model_classes.get(type_name)

    @classmethod
    def __subclasshook__(cls, C):
        if cls.get_model_type_for_class(C) is not None:
            return True
        return NotImplemented


def get_model_class_for_type(model_type, type_name):
    if not hasattr(model_type, 'get_model_class_for_type'):
        return model_type

    model_class = model_type.get_model_class_for_type(type_name)

    if model_class is None:
        raise ValidationError('{} is not in allowed types'.format(type_name))

    return model_class


def get_type_name_for_model(model_type, model_class):
    model_type_name = model_type.get_model_type_for_class(model_class)

    if model_type_name is None:
        raise TypeError('Could not determine model type name for {} using {}'.format(model_class, model_type))

    return model_type_name


def get_type_specifier_name(model_type):
    return getattr(model_type, 'type_specifier_name', None)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
    name='justamodel',
    packages=['justamodel'],
    version='0.0.1',
    description='Simple model schema definition',
    url='https://github.com/martin-sucha/justamodel',
    author='Martin Sucha',
    author_email='sucha14@uniba.sk',
    license='MIT',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)


# -*- coding: utf-8 -*-


class ValidationError(Exception):
    pass


class ModelValidationError(ValidationError):
    def __init__(self, *args, **kwargs):
        super().__init__('Model validation failed', *args, **kwargs)
        self.errors = []
        self.field_errors = {}

    def add_error(self, error):
        self.errors.append(error)

    def get_or_create_field(self, name):
        if name in self.field_errors:
            field_errors = self.field_errors[name]
        else:
            field_errors = ModelValidationError()
            self.field_errors[name] = field_errors
        return field_errors

    def add_field_error(self, name, error):
        if isinstance(error, ModelValidationError):
            if name in self.field_errors:
                raise TypeError('Merging multiple model validation errors is not supported')
            self.field_errors[name] = error
            return

        self.get_or_create_field(name).add_error(error)

    def add_path_error(self, path, error):
        if not path:
            return self.add_error(error)

        spath = path.split('.')
        cur = self
        for part in spath[:-1]:
            cur = cur.get_or_create_field(part)

        cur.add_field_error(spath[-1], error)

    @classmethod
    def for_field(cls, path, error):
        wrapped_error = cls()
        wrapped_error.add_path_error(path, error)
        return wrapped_error

    def get_errors(self, path=''):
        if not path:
            return self.errors

        cur = self
        for part in path.split('.'):
            if part not in cur.field_errors:
                return []

            cur = cur.field_errors[part]

        return cur.errors

    def __bool__(self):
        return bool(self.errors) or bool(self.field_errors)